using System;
using System.Diagnostics;
using System.Windows.Forms;
using Syncfusion.ComponentModel;
using Syncfusion.Windows.Forms.Grid;

namespace EnterKeyDemo
{

    public class ExcelLikeNavigation : Disposable
    {
        GridControl m_grid;
        GridRangeInfo activeRange;

        private delegate bool NextCell(GridModel gridModel, ref GridRangeInfo range, int gridHeight, int gridWidth);

        #region "navigation functions"
        private static bool UpCell(GridModel gridModel, ref GridRangeInfo range, int gridHeight, int gridWidth)
        {
            if (range.Top >= 1)
            {
                do
                {
                    range = range.OffsetRange(-1, 0);
                } while (!gridModel[range.Top, range.Left].Enabled); //skip disabled/readonly cells
                return true;
            }
            return false;
        }
        private static bool DownCell(GridModel gridModel, ref GridRangeInfo range, int gridHeight, int gridWidth)
        {
            if (range.Top < gridHeight)
            {
                do
                {
                    range = range.OffsetRange(1, 0);
                } while (!gridModel[range.Top, range.Left].Enabled);
                return true;
            }
            return false;
        }
        private static bool LeftCell(GridModel gridModel, ref GridRangeInfo range, int gridHeight, int gridWidth)
        {
            if (range.Left >= 1)
            {
                do
                {
                    range = range.OffsetRange(0, -1);
                } while (!gridModel[range.Top, range.Left].Enabled);
                return true;
            }
            return false;
        }
        private static bool RightCell(GridModel gridModel, ref GridRangeInfo range, int gridHeight, int gridWidth)
        {
            if (range.Left < gridWidth)
            {
                do
                {
                    range = range.OffsetRange(0, 1);
                } while (!gridModel[range.Top, range.Left].Enabled);
                return true;
            }
            return false;
        }
        #endregion

        public ExcelLikeNavigation(GridControl grid)
        {
            m_grid = grid;

            m_grid.KeyDown += m_grid_KeyDown;
        }


        void m_grid_KeyDown(object sender, KeyEventArgs e)
        {
            Debug.WriteLine(Enum.GetName(typeof(Keys), e.KeyCode));

            if (!m_grid.CurrentCell.IsEditing && !e.Alt)
            {
                //on cell move - deactivate selections (not selecting)
                if ((e.KeyCode == Keys.Up ||
                     e.KeyCode == Keys.Down ||
                     e.KeyCode == Keys.Left ||
                     e.KeyCode == Keys.Right) && !e.Shift)
                {
                    m_grid.Selections.Clear();
                }

                //check arrow key + shift only
                if ((e.KeyCode == Keys.Up ||
                         e.KeyCode == Keys.Down ||
                         e.KeyCode == Keys.Left ||
                         e.KeyCode == Keys.Right) && e.Shift && !e.Control)
                {
                    m_grid_KeyDownWithShiftOnly(sender, e);
                }

                //now look at combinations of arrow key and ctrl (and shift)
                if (m_grid.CurrentCell != null && e.Control)
                {
                    m_grid_KeyDownWithCtrl(sender, e);
                }

            }

            //****POTENTIAL BUG****
            //This code does not work as the enter key code is not passed in the event arg
            else if (m_grid.CurrentCell.IsEditing && e.Control && !e.Alt && (e.KeyCode == Keys.Enter) && !e.Shift)
            {
                m_grid.CurrentCell.ConfirmChanges();

                if (m_grid.CurrentCell.IsValid && m_grid.Selections.Count > 0)
                {
                    m_grid.CurrentCell.EndEdit();
                    string value = m_grid.Model.GetCellsInfo(m_grid.CurrentCell.RangeInfo)[0].Text;
                    GridRangeInfo range = m_grid.Selections.Ranges[0];
                    for (int row = range.Top; row <= range.Bottom; row++)
                    {
                        for (int col = range.Left; col <= range.Right; col++)
                        {
                            if (!m_grid[row, col].ReadOnly)
                            {
                                m_grid[row, col].Text = value;
                            }
                        }
                    }
                }
            }

        }

        void m_grid_KeyDownWithShiftOnly(object sender, KeyEventArgs e)
        {
            GridRangeInfo origin;
            if (m_grid.Selections.Count > 0)
            {
                int row;
                int col;
                if (m_grid.CurrentCell.RowIndex == m_grid.Selections.Ranges[0].Top)
                {
                    row = m_grid.Selections.Ranges[0].Bottom;
                }
                else
                {
                    row = m_grid.Selections.Ranges[0].Top;
                }
                if (m_grid.CurrentCell.ColIndex == m_grid.Selections.Ranges[0].Left)
                {
                    col = m_grid.Selections.Ranges[0].Right;
                }
                else
                {
                    col = m_grid.Selections.Ranges[0].Left;
                }
                origin = GridRangeInfo.Cell(row, col);
            }
            else
            {
                origin = m_grid.CurrentCell.RangeInfo;
            }
            switch (e.KeyCode)
            {
                case Keys.Up:
                    m_grid.CurrentCell.MoveUp(1);
                    break;
                case Keys.Down:
                    m_grid.CurrentCell.MoveDown(1);
                    break;
                case Keys.Left:
                    m_grid.CurrentCell.MoveLeft(1);
                    break;
                case Keys.Right:
                    m_grid.CurrentCell.MoveRight(1);
                    break;
                default:
                    break;
            }
            ExtendSelection(m_grid.CurrentCell, origin);
            e.SuppressKeyPress = true;
        }
        void m_grid_KeyDownWithCtrl(object sender, KeyEventArgs e)
        {
            activeRange = m_grid.CurrentCell.RangeInfo;
            if (activeRange != null && !activeRange.IsEmpty)
            {
                //work on the top left cell
                GridRangeInfo testCell = GridRangeInfo.Cells(activeRange.Top, activeRange.Left, activeRange.Top, activeRange.Left);
                NextCell cellNavFunction = getCellNav(e.KeyCode);

                if (cellNavFunction != null)
                {
                    bool latchEmpty = m_grid[testCell.Top, testCell.Left].IsEmpty
                                        || (m_grid[testCell.Top, testCell.Left].CellModel is GridPushButtonCellModel)
                                        || m_grid.Model.HideRows[testCell.Top]
                                        || m_grid.Model.HideCols[testCell.Left];
                    //treat push button and not visible as empty
                    int prevRow = -1;
                    int prevCol = -1;
                    bool firstPass = true;
                    while (cellNavFunction(m_grid.Model, ref testCell, m_grid.RowCount, m_grid.ColCount))
                    {
                        //stop only when end/beginning is reached or empty changes (latch)
                        if (latchEmpty !=
                            (m_grid[testCell.Top, testCell.Left].IsEmpty ||
                            (m_grid[testCell.Top, testCell.Left].CellModel is GridPushButtonCellModel) ||
                            m_grid.Model.HideRows[testCell.Top] ||
                            m_grid.Model.HideCols[testCell.Left]
                            ))
                        //treat push button, and hidden as empty
                        {
                            break;
                        }
                        //Also stop if we are at the top or left hand side
                        else if ((testCell.Top == 0) || (testCell.Left == 0))
                        {
                            break;
                        }
                        //only allow selection of cells that can be selected
                        prevRow = testCell.Top;
                        prevCol = testCell.Left;
                        firstPass = false;
                    }
                    int rowOffset;
                    int colOffset;
                    if (firstPass)
                    {
                        rowOffset = testCell.Top - activeRange.Top;
                        colOffset = testCell.Left - activeRange.Left;
                    }
                    else
                    {
                        rowOffset = prevRow - activeRange.Top;
                        colOffset = prevCol - activeRange.Left;
                    }
                    //actualy do the move (or selection)
                    MoveOrSelectToEnabled(m_grid, rowOffset, colOffset, e.Shift);
                    e.SuppressKeyPress = true;
                }
            }
        }

        private NextCell getCellNav(Keys keyCode)
        {
            switch (keyCode)
            {
                case Keys.Up:
                    return ExcelLikeNavigation.UpCell;
                case Keys.Down:
                    return ExcelLikeNavigation.DownCell;
                case Keys.Left:
                    return ExcelLikeNavigation.LeftCell;
                case Keys.Right:
                    return ExcelLikeNavigation.RightCell;
                default:
                    return null;
            }
        }

        private void MoveOrSelectToEnabled(GridControl grid, int rowOffset, int colOffset, bool select)
        {
            bool done = false;
            int rowTryOffset = rowOffset;
            int colTryOffset = colOffset;
            GridCurrentCell currentCell = grid.CurrentCell;
            int rowOrigin = currentCell.RangeInfo.Top;
            int colOrigin = currentCell.RangeInfo.Left;
            int targetRow = rowOrigin + rowTryOffset;
            int targetCol = colOrigin + colTryOffset;
            int maxRow = grid.RowCount;
            int maxCol = grid.ColCount;
            int increment;
            GridRangeInfo origin = currentCell.RangeInfo;

            //origin must be set to last selection if a selection is present
            if ((grid.Selections.Ranges.Count > 0) &&
                (grid.Selections.Ranges[grid.Selections.Ranges.Count - 1].Contains(currentCell.RangeInfo)))
            {
                origin = grid.Selections.Ranges[grid.Selections.Ranges.Count - 1];
            }

            if ((rowTryOffset > 1) || (colTryOffset > 1) || (rowTryOffset < -1) || (colTryOffset < -1))
            {
                increment = -1;
            }
            else
            {
                increment = 1;
            }


            while (!done && (targetRow <= maxRow) && (targetCol <= maxCol) &&
                            (targetRow >= 1) && (targetCol >= 1))
            {
                done = currentCell.MoveTo(targetRow, targetCol, GridSetCurrentCellOptions.None, false);
                //try until we hit an enabled cell
                if (!done)
                {
                    if (rowTryOffset < 0)
                    {
                        rowTryOffset = rowTryOffset - increment;
                    }
                    else if (rowTryOffset > 0)
                    {
                        rowTryOffset = rowTryOffset + increment;
                    }

                    if (colTryOffset < 0)
                    {
                        colTryOffset = colTryOffset - increment;
                    }
                    else if (colTryOffset > 0)
                    {
                        colTryOffset = colTryOffset + increment;
                    }

                    if ((colTryOffset == 0) && (rowTryOffset == 0))
                    {
                        if ((increment == -1) && (rowTryOffset == 0) && (colTryOffset == 0))
                        {
                            //-1 failed - try +1
                            increment = 1;
                            //reset offsets
                            rowTryOffset = rowOffset;
                            colTryOffset = colOffset;
                        }
                        else
                        {
                            //fail
                            break;
                        }
                    }

                    //reset target row and col
                    targetRow = rowOrigin + rowTryOffset;
                    targetCol = colOrigin + colTryOffset;
                }
            }

            if (select)
            {
                //select the range
                ExtendSelection(currentCell, origin);
            }

            if (!grid.CurrentCell.IsVisible)
            {
                grid.CurrentCell.ScrollInView();
            }
        }

        /// <summary>
        /// Moves specified rows and columns extending selection.
        /// Usualy this will only be used to extend one dimension at a time.
        /// </summary>
        /// <param name="currentCell">Current cell selection</param>
        /// <param name="rowMove">Number of rows to move - = up</param>
        /// <param name="colMove">Number of cols to move - = left</param>
        /// <returns></returns>
        private void ExtendSelection(GridCurrentCell newCurrentCell, GridRangeInfo origin)
        {
            int top = Math.Min(newCurrentCell.RowIndex, origin.Top);
            int left = Math.Min(newCurrentCell.ColIndex, origin.Left);
            int bottom = Math.Max(newCurrentCell.RowIndex, origin.Bottom);
            int right = Math.Max(newCurrentCell.ColIndex, origin.Right);
            newCurrentCell.Grid.Selections.Clear();
            newCurrentCell.Grid.Selections.Add(GridRangeInfo.Cells(top, left, bottom, right));
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                m_grid.KeyDown -= m_grid_KeyDown;
            }
            base.Dispose(disposing);
        }
    }
}
